import uuid
import aiohttp

from .models import Action, Bot, BotPost, List, ListUpdate

BASE_URL = "https://catnip.metrobots.xyz"
"""The base URL for the Metro API."""

class MetroHTTP():
    """Contains the HTTP methods for the Metro Reviews API"""
    def __init__(self, *, list_id: str, secret_key: str):
        self.list_id = list_id
        self.secret_key = secret_key

    async def request(self, *, url: str, method: str, json: dict = None):
        """Make a request to the Metro Reviews API"""
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method, 
                BASE_URL+url, 
                json=json, 
                headers={
                    "Authorization": self.secret_key
                }
            ) as resp:
                resp_data = await resp.json()
                return resp, resp_data

    async def update_list(self, update: ListUpdate):
        """Edits the list"""
        return await self.request(
            url=f"/lists/{self.list_id}",
            method="PATCH",
            json=update.dict()
        )
    
    async def add_bot(self, bot: BotPost):
        """Add a bot to Metro Reviews"""
        res, json = await self.request(url=f"/bots?list_id={self.list_id}", method="POST", json=bot.dict())

        if not res.ok:
            raise RuntimeError(f"Failed to add bot: {res.status}")

        return Bot(**json)

    async def approve_bot(self, *, bot_id: str, reviewer: str, reason: str):
        """Approve a bot"""
        return await self.request(
            url=f"/bots/{bot_id}/approve?reviewer={reviewer}&list_id={self.list_id}",
            method="POST",
            json={
                "reason": reason
            }
        )

    async def deny_bot(self, *, bot_id: str, reviewer: str, reason: str):
        """Dnies a bot"""
        return await self.request(
            url=f"/bots/{bot_id}/deny?reviewer={reviewer}&list_id={self.list_id}",
            method="POST",
            json={
                "reason": reason
            }
        )
    
    async def get_actions(self, *, offset: int, limit: int):
        """
**This is a paginated API**

Get actions from Metro Reviews. 

You need to use ``Metro.paginate`` with this as this endpoint is paginated.
        """
        res, json = await self.request(
            url=f"/actions?offset={offset}&limit={limit}",
            method="GET"
        )

        if not res.ok:
            raise RuntimeError(f"Failed to get actions: {res.status}")

        return json, Action

    async def get_all_lists(self) -> list[List]:
        """Get all lists from Metro Reviews"""
        res, json = await self.request(url="/lists", method="GET")
        
        if not res.ok:
            raise RuntimeError(f"Failed to get lists: {res.status}")
        
        return [List(**i) for i in json]

    async def get_list(self, list_id: str | uuid.UUID) -> List:
        """Get a list from Metro Reviews"""
        res, json = await self.request(url=f"/list/{list_id}", method="GET")

        if not res.ok:
            raise RuntimeError(f"Failed to get list: {res.status}")

        return List(**json)

    async def get_all_bots(self) -> list[Bot]:
        """Get all bots from Metro Reviews"""
        res, json = await self.request(url=f"/bots", method="GET")

        if not res.ok:
            raise RuntimeError(f"Failed to get bots: {res.status}")

        return [Bot(**i) for i in json]
    
    async def get_bot(self, bot_id: str | uuid.UUID) -> Bot:
        """Get a bot from Metro Reviews"""
        res, json = await self.request(url=f"/bots/{bot_id}", method="GET")

        if not res.ok:
            raise RuntimeError(f"Failed to get bot: {res.status}")

        return Bot(**json)



from functools import wraps
from typing import Awaitable
from fastapi import FastAPI, Request, Response
from pydantic import BaseModel

from .http import MetroHTTP
from .models import Bot, ListUpdate

DEFAULT_TAGS = ["Metro (Integrase)"]
"""Default tags for the integrase routes"""

class Metro():
    """
This is the actual main client you should initialize and then use

To use the server bit, you should do something like this:

```py

from metro_integrase import Metro
from fastapi import FastAPI, Request

app = FastAPI()
metro = Metro(...)

@metro.claim()
async def claim(bot: Bot):
    ...

@metro.unclaim()
async def unclaim(bot: Bot):
    ...

@metro.approve()
async def approve(bot: Bot):
    ...

@metro.deny()
async def deny(bot: Bot):
    ...

@app.on_event("startup")
async def startup():
    await metro.register_api_urls()
```

If you wish to use a HTTP method, use ``self.http`` or manually initialize ``MetroHTTP`` with the list ID and secret key.
    """

    def __init__(self, *, domain: str, list_id: str, secret_key: str, app: FastAPI = None):
        self.http = MetroHTTP(list_id=list_id, secret_key=secret_key)
        self.domain = domain
        self._urls = {}
        self._app = app
    
    async def paginate(self, func: Awaitable, *, limit: int = 50):
        """
Simple helper that can be combined with a ``self.http`` paginated function to paginate a function

Example:

```py
    async for act in metro.paginate(metro.http.get_actions):
        print(act)
```
"""
        offset = 0
        while True:
            ret = await func(offset=offset, limit=limit)

            json: list[dict] = ret[0]
            model: BaseModel = ret[1]

            if not json:
                break
            for item in json:
                yield model(**item)
            offset += limit

    async def register_api_urls(self):
        """Register all decorated API endpoints with Metro Reviews using ``MetroHTTP.update_list``"""
        res, json = await self.http.update_list(
            ListUpdate(
                domain=self.domain,
                claim_bot_api=self._urls.get("claim"),
                unclaim_bot_api=self._urls.get("unclaim"),
                approve_bot_api=self._urls.get("approve"),
                deny_bot_api=self._urls.get("deny")
            )
        )

        if not res.ok:
            try:
                if json["detail"][0]["loc"][1] == "list_id":
                    raise ValueError("Invalid list ID")
            except (KeyError, IndexError):
                ...

            raise RuntimeError(f"Metro setup failed: {res.status}. Ensure proper list ID and secret key...")
    
    # HTTP Server Code
    def _wrapper(self, *, url: str, name: str, func: Awaitable, tags: list[str]):
        if not self._app:
            raise ValueError("App must be passed in order to use this method")

        @wraps(func)
        async def metro_f(request: Request, response: Response, bot: Bot, *args, **kwargs):
            if request.headers.get("Authorization") != self.http.secret_key:
                response.status_code = 401
                return {"detail": "Invalid secret key"}

            return await func(bot, *args, **kwargs)

        self._app.post(url, tags=tags)(metro_f)

        self._urls[name] = url

    def claim(self, *, tags: list[str] = DEFAULT_TAGS, url: str = "/metro/claim"):
        """"Claim API Decorator"""

        def wrapper(func: Awaitable):
            self._wrapper(url=url, name="claim", func=func, tags=tags)
        
        return wrapper

    def unclaim(self, *, tags: list[str] = DEFAULT_TAGS, url: str = "/metro/unclaim"):
        """"Unclaim API Decorator"""

        def wrapper(func: Awaitable):
            self._wrapper(url=url, name="unclaim", func=func, tags=tags)
        
        return wrapper

    def approve(self, *, tags: list[str] = DEFAULT_TAGS, url: str = "/metro/approve"):
        """"Approve API Decorator"""

        def wrapper(func: Awaitable):
            self._wrapper(url=url, name="approve", func=func, tags=tags)
        
        return wrapper

    def deny(self, *, tags: list[str] = DEFAULT_TAGS, url: str = "/metro/deny"):
        """"Deny API Decorator"""

        def wrapper(func: Awaitable):
            self._wrapper(url=url, name="deny", func=func, tags=tags)
        
        return wrapper
    



import datetime
import enum
import uuid
from pydantic import BaseModel

# https://github.com/MetroReviews/backend/blob/74c902cc8a10e84796ad779228ea5417fe6ba087/brc/tables.py#L9
class ActionEnum(enum.IntEnum):
    """
Type of action
    """
    CLAIM = 0
    UNCLAIM = 1
    APPROVE = 2
    DENY = 3

class State(enum.IntEnum):
    """
Current bot state
    """
    PENDING = 0
    UNDER_REVIEW = 1
    APPROVED = 2
    DENIED = 3

class ListState(enum.IntEnum):
    """
A lists state
    """
    PENDING_API_SUPPORT = 0
    SUPPORTED = 1
    DEFUNCT = 2
    BLACKLISTED = 3
    UNCONFIRMED_ENROLLMENT = 4

# https://github.com/MetroReviews/backend/blob/64959ddaa0faecfc38007580cc1625412b9b5864/brc/app.py#L109
class List(BaseModel):
    id: uuid.UUID
    name: str
    description: str | None = None
    domain: str | None = None
    state: ListState
    icon: str | None = None


# https://github.com/MetroReviews/backend/blob/74c902cc8a10e84796ad779228ea5417fe6ba087/brc/app.py#L117
class BotPost(BaseModel):
    bot_id: str
    banner: str | None = None
    description: str 
    long_description: str
    website: str | None = None
    invite: str | None = None
    owner: str
    extra_owners: list[str] | None = []
    support: str | None = None
    donate: str | None = None
    library: str | None = None
    nsfw: bool | None = False
    prefix: str | None = None
    tags: list[str] | None = None
    review_note: str | None = None
    cross_add: bool | None = True


class Bot(BotPost):
    state: State
    list_source: uuid.UUID
    added_at: datetime.datetime
    reviewer: str | None = None
    invite_link: str | None = None
    username: str | None = "Unknown"

class ListUpdate(BaseModel):
    name: str | None = None
    description: str | None = None
    domain: str | None = None
    claim_bot_api: str | None = None
    unclaim_bot_api: str | None = None
    approve_bot_api: str | None = None
    deny_bot_api: str | None = None
    reset_secret_key: bool = False
    icon: str | None = None

# https://github.com/MetroReviews/backend/blob/main/brc/app.py#L247
class Action(BaseModel):
    id: int
    bot_id: str
    action: ActionEnum
    reason: str
    reviewer: str 
    action_time: datetime.datetime
    list_source: uuid.UUID



import setuptools 

setuptools.setup(
    name="metro_integrase",
    version="0.0.2",
    author="Rootspring",
    author_email="cheesycod@outlook.com",
    description="A Python wrapper for the Metro Reviews API",
    long_description="A Python wrapper for the Metro Reviews API (https://github.com/MetroReviews)",
    include_dirs=["metro_integrase"],
    install_requires=[
        "fastapi",
        "aiohttp",
        "pydantic",
    ]
)